# -*- coding: UTF-8 -*-
"""
This module provides an abstract base class for storing logging or property information.
"""

import sqlalchemy as sq
import sys

from geological_toolbox.db_handler import AbstractDBObject


class AbstractLogClass(AbstractDBObject):
    """
    This class is the base for storing logging information or properties. It should be treated as abstract, no
    object should be created directly!

    :param property_name: name of the log or property
    :param property_unit: unit of the log values or properties
    :return: Nothing
    """

    # define table_columns
    prop_name = sq.Column(sq.VARCHAR(50), default="")
    prop_unit = sq.Column(sq.VARCHAR(100), default="")

    def __init__(self, property_name: str, property_unit: str, *args, **kwargs) -> None:
        """
        Initialises the class
        """

        self.property_name = property_name
        self.property_unit = property_unit

        AbstractDBObject.__init__(self, *args, **kwargs)

    def __repr__(self):
        text = "<AbstractLogClass property_name={}, property_unit={}\n".\
            format(self.property_name, self.property_unit)
        text += AbstractDBObject.__repr__(self)
        return text

    def __str__(self):
        text = "{} [{}]\n".format(self.property_name, self.property_unit)
        text += AbstractDBObject.__str__(self)
        return text

    @property
    def property_name(self) -> str:
        """
        name of the property

        :return: the name of the property
        """
        return str(self.prop_name)

    @property_name.setter
    def property_name(self, name: str) -> None:
        """
        name of the property

        :param name: new property name
        :return: Nothing
        """
        name = str(name)
        if len(name) > 100:
            name = name[:100]
        self.prop_name = name

    @property
    def property_unit(self) -> str:
        """
        property unit as string

        :return: the unit of the property
        """
        if sys.version_info[0] == 2:
            return self.prop_unit
        else:
            return self.prop_unit

    @property_unit.setter
    def property_unit(self, unit: str) -> None:
        """
        property unit as string

        :param unit: property unit as string
        :return: Nothing
        """
        if sys.version_info[0] == 2:
            unit = unit

        if len(unit) > 100:
            unit = unit[:100]
        self.prop_unit = unit





# -*- coding: UTF-8 -*-
"""
This module provides a class for storing stratigraphical information in a database.
"""

import sqlalchemy as sq
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.session import Session
from typing import List

from geological_toolbox.exceptions import DatabaseException
from geological_toolbox.db_handler import Base, AbstractDBObject


class StratigraphicObject(Base, AbstractDBObject):
    """
    A class for storing stratigraphical information in a database.

    :param name: Name of the stratigraphic unit
    :param age: age of the stratigraphic unit (-1 if none)
    :return: nothing
    :raises ValueError: if age is not compatible to float
    """
    # define db table name and columns
    __tablename__ = "stratigraphy"

    id = sq.Column(sq.INTEGER, sq.Sequence("stratigraphy_id_seq"), primary_key=True)
    unit_name = sq.Column(sq.VARCHAR(50), unique=True)
    age = sq.Column(sq.FLOAT(), default=-1)

    def __init__(self, name: str, age: float = -1, *args, **kwargs) -> None:
        """
        Initialize a stratigraphic unit
        """
        AbstractDBObject.__init__(self, *args, **kwargs)

        try:
            age = int(age)
        except ValueError as e:
            raise ValueError("Cannot convert age to int:\n{}".format(str(e)))

        self.unit_name = str(name)
        self.age = -1 if (age < 0) else age

    def __repr__(self) -> str:
        return "<AbstractDBObject(id='{}', name='{}', age='{}')>".format(self.id, self.unit_name, self.age)

    def __str__(self) -> str:
        return "[{}]: name='{}', age='{}'".format(self.id, self.unit_name, self.age)

    # define setter and getter for columns and local data
    @property
    def horizon_age(self) -> float:
        """
        The age of the stratigraphic unit

        :raises ValueError: if new value is convertible to type int
        """
        return float(self.age)

    @horizon_age.setter
    def horizon_age(self, value: float) -> None:
        """
        See getter
        """
        try:
            value = float(value)
        except ValueError:
            raise ValueError("Cannot convert value ({}) to type int.".format(str(value)))

        if value <= 0:
            self.age = -1
        else:
            self.age = value

    @property
    def statigraphic_name(self) -> str:
        """
        The statigraphic_name of the stratigraphic unit
        """
        return self.unit_name

    @statigraphic_name.setter
    def statigraphic_name(self, value: str) -> None:
        """
        see getter
        """
        self.unit_name = str(value)

    @classmethod
    def init_stratigraphy(cls, session: Session, name: str, age: float = -1,
                          update: bool = False) -> "StratigraphicObject":
        """
        Initialize a stratigraphic unit. Create a new one if unit doesn't exists in the database, else use the existing.

        :param session: SQLAlchemy session, which includes the database connection
        :param name: Name of the stratigraphic unit
        :param age: age of the stratigraphic unit (-1 if none)
        :param update: update age if stratigraphic unit exists
        :raises ValueError: if age is not compatible to float
        :raises TypeError: if session is not of type SQLAlchemy Session
        """
        if not isinstance(session, Session):
            raise TypeError("'session' is not of type SQLAlchemy Session!")

        try:
            age = float(age)
        except ValueError as e:
            raise ValueError("Cannot convert age to float:\n{}".format(str(e)))

        if age < 0:
            age = -1

        # check if horizon exists (unique name)
        result = session.query(StratigraphicObject).filter(StratigraphicObject.unit_name == name)
        if result.count() == 0:  # no result found -> create new stratigraphic unit
            return cls(name, age, session=session)
        if result.count() == 1:  # one result found -> stratigr. unit exists in db -> return and possibly update
            result = result.one()
            result.session = session
            if update:  # change age value
                result.horizon_age = age
            return result

        # more than one result? => heavy failure, statigraphic_name should be unique => DatabaseException
        # for res in result.all():
        #     print(res)
        raise DatabaseException(
            "More than one ({}) horizon with the same statigraphic_name: {}! Database error!".format(result.count(), name))

    # load units from db
    @classmethod
    def load_all_from_db(cls, session: Session) -> List["StratigraphicObject"]:
        """
        Returns all stratigraphic units stored in the database connected to the SQLAlchemy Session session

        :param session: represents the database connection as SQLAlchemy Session
        :return: a list of stratigraphic units representing the result of the database query
        :raises TypeError: if session is not of type SQLAlchemy Session
        """
        if not isinstance(session, Session):
            raise TypeError("'session' is not of type SQLAlchemy Session!")

        result = session.query(cls).all()
        for horizon in result:  # set session value
            horizon.session = session
        return result

    @classmethod
    def load_by_stratigraphic_name_from_db(cls, name: str, session: Session) -> "StratigraphicObject" or None:
        """
        Returns the stratigraphic unit with the given name in the database connected to the SQLAlchemy Session session

        :param name: The name of the requested stratigraphic unit
        :param session: represents the database connection as SQLAlchemy Session
        :return: As the name is a unique value, only one result can be returned or None
        :raises DatabaseException: if more than one result was found (name is an unique value)
        :raises TypeError: if session is not of type SQLAlchemy Session
        """
        if not isinstance(session, Session):
            raise TypeError("'session' is not of type SQLAlchemy Session!")

        result = session.query(cls).filter(cls.unit_name == name)
        if result.count() == 0:
            return None
        if result.count() == 1:
            result = result.one()
            result.session = session
            return result

        raise DatabaseException("More than one ({}) horizon with the same statigraphic_name: {}! Database error!".
                                format(result.count(), name))

    @classmethod
    def load_by_age_from_db(cls, min_age: float, max_age: float, session: Session) -> List["StratigraphicObject"]:
        """
        Returns a list of stratigraphic units with an age between min_age and max_age from the database connected to
        the SQLAlchemy Session session. If no result was found, this function returns an empty list.

        :param min_age: Minimal age of the stratigraphic units
        :param max_age: Maximal age of the stratigraphic units
        :param session: represents the database connection as SQLAlchemy Session
        :return: Returns a list of stratigraphic units with an age between min_age and max_age.
        :raises ValueError: if min_age or max_age is not compatible to float
        :raises TypeError: if session is not of type SQLAlchemy Session
        """
        min_age = float(min_age)
        max_age = float(max_age)
        if not isinstance(session, Session):
            raise TypeError("'session' is not of type SQLAlchemy Session!")

        result = session.query(cls).filter(sq.between(cls.age, min_age, max_age)).all()
        for horizon in result:
            horizon.session = session
        return result



# -*- coding: UTF-8 -*-
"""
This module provides a class for storing logging information of wells. The logging class is derived from the
AbstractLogClass.
"""

import sqlalchemy as sq
from sqlalchemy.orm import relationship
from typing import List

from geological_toolbox.abstract_log import AbstractLogClass
from geological_toolbox.db_handler import Base, AbstractDBObject


class WellLogValue(Base, AbstractDBObject):
    """
    This class connects the logging values to their related depth

    :param depth: depth of the log value
    :param value: log value at the specified depth
    :param args: parameters for AbstractDBObject initialisation
    :param kwargs: parameters for AbstractDBObject initialisation
    :return: Nothing
    :raises ValueError: if a parameter has wrong type or cannot be converted to required
    """
    # define db table stratigraphic_name and columns
    __tablename__ = "logging_association"

    id = sq.Column(sq.INTEGER, sq.Sequence("logging_association_id_seq"), primary_key=True)
    log_depth = sq.Column(sq.FLOAT)
    log_value = sq.Column(sq.FLOAT)

    log_id = sq.Column(sq.INTEGER, sq.ForeignKey("well_logs.id"), default=-1)
    sq.Index("welllogdepth_index", log_depth)

    def __init__(self, depth: float, value: float, *args, **kwargs) -> None:
        """
        Initialise the class
        """
        self.depth = depth
        self.value = value

        AbstractDBObject.__init__(self, *args, **kwargs)

    def __repr__(self) -> str:
        text = "<WellLogValue(id='{}', depth='{}', value='{}', log_id='{}'),\n". \
            format(self.id, self.depth, self.value, self.log_id)
        text += "Additional DBObject: {}>".format(AbstractDBObject.__repr__(self))
        return text

    def __str__(self) -> str:
        text = "[{}] {}: {} (well_log.id: {})\n" \
            .format(self.id, self.depth, self.value, self.log_id)
        text += "DBObject: {}".format(AbstractDBObject.__str__(self))
        return text

    @property
    def depth(self) -> float:
        """
        depth of the log value

        :raises ValueError: if log_depth is not type float or cannot be converted to it
        """
        return float(self.log_depth)

    @depth.setter
    def depth(self, log_depth: float) -> None:
        """
        see getter
        """
        self.log_depth = float(log_depth)

    @property
    def value(self) -> float:
        """
        log value

        :raises ValueError: Raises ValueError if log_value is not type float or cannot be converted to it
        """
        return float(self.log_value)

    @value.setter
    def value(self, log_value: float) -> None:
        """
        see getter
        """
        self.log_value = float(log_value)


class WellLog(Base, AbstractLogClass):
    """
    represents logging information for wells
    """
    # define db table well_logs and columns
    __tablename__ = "well_logs"

    id = sq.Column(sq.INTEGER, sq.Sequence("well_logs_id_seq"), primary_key=True)
    well_id = sq.Column(sq.INTEGER, sq.ForeignKey("wells.id"), default=-1)
    # define markers relationship
    log_values = relationship("WellLogValue", order_by=WellLogValue.log_depth,
                              backref="well_log", primaryjoin="WellLog.id==WellLogValue.log_id",
                              cascade="all, delete, delete-orphan")

    def __init__(self, *args, **kwargs) -> None:
        """
        Initialise the class
        """
        AbstractLogClass.__init__(self, *args, **kwargs)

    def __repr__(self) -> str:
        text = "<WellLog(id='{}', well_id='{}', log values='{}')>\n". \
            format(self.id, self.well.id, str(self.log_values))
        text += "Additional DBObject: {}".format(AbstractDBObject.__repr__(self))
        return text

    def __str__(self) -> str:
        text = "[{}] Well ({}: {})\n" \
            .format(self.id, self.well.id, self.well.name)
        text += "DBObject: {}".format(AbstractDBObject.__str__(self))
        for assoc in self.log_values:
            text += "{}".format(str(assoc))
        return text

    def insert_log_value(self, log_value: WellLogValue) -> None:
        """
        Insert a new log value in the log
        ATTENTION: If you insert a log value, the log will be automatically stored in the database!

        :param log_value: WellLogValue to be inserted
        :return: Nothing
        :raises TypeError: if marker is not an instance of WellLogValue
        :raises ValueError: if the depth of the marker is larger than the drilled depth of the well or < 0
        """
        if not isinstance(log_value, WellLogValue):
            raise TypeError("log_value {} is not of type WellLogValue!".format(str(log_value)))
        if log_value.depth < 0:
            raise ValueError("Value depth ({}) < 0!".format(log_value.depth))
        if (not (self.well is None)) and (log_value.depth > self.well.depth):
            raise ValueError("Value depth ({}) is larger than final well depth ({})!".format(log_value.depth,
                                                                                             self.well.depth))
        self.log_values.append(log_value)

        # new sorting to ensure correct order without storage and reloading from the database
        self.log_values.sort(key=lambda x: x.depth)

    def insert_multiple_log_values(self, log_values: List[WellLogValue]) -> None:
        """
        Insert the multiple log values in the log
        ATTENTION: If you insert values, the log will be automatically stored in the database!

        :param log_values: List of marker to be inserted
        :return: Nothing
        :raises TypeError: if one of the marker is not an instance of WellLogValue
        :raises ValueError: if the depth of a value is larger than the drilled depth of the well or < 0
        """
        for value in log_values:
            if not isinstance(value, WellLogValue):
                raise TypeError(
                    "At least on value is not of type WellLogValue ({}: {})!".format(str(value), str(type(value))))
            if value.depth < 0:
                raise ValueError("Value depth ({}) < 0!".format(value.depth))
            if (not (self.well is None)) and (value.depth > self.well.depth):
                raise ValueError("Value depth ({}) is larger than final well depth ({})!".
                                 format(value.depth, self.well.depth))

        self.log_values += log_values

        # new sorting to ensure correct order without storage and reloading from the database
        self.log_values.sort(key=lambda x: x.depth)

    def get_value_by_depth(self, depth: float) -> WellLogValue or None:
        """
        Returns the value at depth "depth"

        :param depth: depth of the requested marker
        :return: Returns the value at depth "depth"
        :raises ValueError: if no marker was found for the committed depth or depth is not compatible to float
        """
        depth = float(depth)
        for value in self.log_values:
            if value.depth == depth:
                return value
        raise ValueError("No value found at depth {}".format(depth))

    def delete_value(self, value: WellLogValue) -> None:
        """
        Deletes the value from the well log

        :param value: WellLogValue object which should be deleted
        :return: Nothing
        :raises TypeError: if value is not of type WellLogValue
        :raises ValueError: if the value is not part of the well log
        """
        if not isinstance(value, WellLogValue):
            raise TypeError("marker {} is not an instance of WellLogValue!".format(str(value)))

        try:
            self.log_values.remove(value)
        except ValueError as e:
            raise ValueError("{}\nWellLogValue with ID {} not found in value list!".format(str(e), value.id))



# -*- coding: UTF-8 -*-
"""
This module hosts the basic AbstractGeoObject class. In declares all basic functions for a GeoObjects
"""

import sqlalchemy as sq
from sqlalchemy.orm.session import Session
from typing import List

from geological_toolbox.db_handler import AbstractDBObject


class AbstractGeoObject(AbstractDBObject):
    """
    This is a base class which stores central GeoObject data. This class should be treated as abstract, no object
    should be created directly!

    :param reference_system: Stores the spatial reference information in a string format (e.g. WKT). This parameter will
           not be verified!
    :param easting: easting coordinate of the object
    :param northing: northing coordinate of the object
    :param altitude: height above sea level of the object or None
    :param args: parameters for AbstractDBObject initialisation
    :param kwargs: parameters for AbstractDBObject initialisation
    :return: nothing
    """

    east = sq.Column(sq.FLOAT)
    north = sq.Column(sq.FLOAT)
    alt = sq.Column(sq.FLOAT)
    reference = sq.Column(sq.TEXT, default="")

    def __init__(self, reference_system: str, easting: float, northing: float, altitude: float,
                 *args, **kwargs) -> None:
        """
        Initialise the AbstractGeoObject
        """

        self.reference_system = reference_system
        self.easting = easting
        self.northing = northing
        self.altitude = altitude

        # call constructor of base class
        AbstractDBObject.__init__(self, *args, **kwargs)

    def __repr__(self) -> str:
        text = "<AbstractGeoObject(east='{}', north='{}', alt='{}')>\n".format(self.easting, self.northing,
                                                                               self.altitude)
        text += AbstractDBObject.__repr__(self)
        return text

    def __str__(self) -> str:
        text = "{} - {} - {} - ".format(self.easting, self.northing, self.altitude)
        text += AbstractDBObject.__str__(self)
        return text

    # define setter and getter for columns and local data
    @property
    def easting(self) -> float:
        """
        The easting value of the object

        :raises ValueError: if value is not compatible to type float
        """
        return float(self.east)

    @easting.setter
    def easting(self, value: float) -> None:
        """
        see getter
        """
        self.east = float(value)

    @property
    def northing(self) -> float:
        """
        The northing value of the object

        :raises ValueError: if value is not compatible to type float
        """
        return float(self.north)

    @northing.setter
    def northing(self, value: float) -> None:
        """
        see getter
        """
        self.north = float(value)

    @property
    def altitude(self) -> float:
        """
        The height above sea level of the object

        :raises ValueError: if value is not compatible to type float
        """
        return float(self.alt)

    @altitude.setter
    def altitude(self, value: float) -> None:
        """
        see getter
        """
        self.alt = float(value)

    @property
    def reference_system(self) -> str:
        """
        The reference system in string format (e.g. WKT)
        ATTENTION: The reference system won't be verified

        :return: Returns the current reference system
        """
        return self.reference

    @reference_system.setter
    def reference_system(self, reference: str) -> None:
        """
        see getter
        """
        self.reference = str(reference)

    @classmethod
    def load_in_extent_from_db(cls, session: Session, min_easting: float, max_easting: float, min_northing: float,
                               max_northing: float) -> List["AbstractGeoObject"]:
        """
        Returns all objects inside the given extent in the database connected to the SQLAlchemy Session session.

        :param session: represents the database connection as SQLAlchemy Session
        :param min_easting: minimal easting of extent
        :param max_easting: maximal easting of extent
        :param min_northing: minimal northing of extent
        :param max_northing: maximal northing of extent
        :return: a list of objects representing the result of the database query
        :raises ValueError: if one of the extension values is not compatible to type float
        :raises TypeError: if session is not of type SQLAlchemy Session
        """
        min_easting = float(min_easting)
        max_easting = float(max_easting)
        min_northing = float(min_northing)
        max_northing = float(max_northing)

        if not isinstance(session, Session):
            raise TypeError("'session' is not of type SQLAlchemy Session!")

        result = session.query(cls).filter(sq.between(cls.east, min_easting, max_easting)). \
            filter(sq.between(cls.north, min_northing, max_northing))
        result = result.order_by(cls.id).all()
        for obj in result:
            obj.session = session
        return result



# -*- coding: UTF-8 -*-
"""
Module providing package wide Exceptions
"""


class ArgumentError(Exception):
    """
    Should be raised, if a wrong number of arguments are submitted or an other failure inside the arguments are
    recognised.

    :param msg: detailed error message
    """

    def __init__(self, msg: str) -> None:
        self.message = str(msg)

    def __str__(self) -> str:
        return self.message


class DatabaseException(Exception):
    """
    Should be raised, if an unresolved database issue occurred (e.g. more than one value in a unique column)

    :param msg: detailed error message
    """

    def __init__(self, msg: str) -> None:
        self.message = str(msg)

    def __str__(self) -> str:
        return self.message


class DatabaseRequestException(Exception):
    """
    Should be raised, if an error occurs during a database request (e.g. not the expected result)

    :param msg: detailed error message
    """

    def __init__(self, msg: str) -> None:
        self.message = str(msg)

    def __str__(self) -> str:
        return self.message


class FaultException(Exception):
    """
    Should be raised, if a fault marker causes an interruption

    :param msg: detailed error message
    """

    def __init__(self, msg: str) -> None:
        self.message = str(msg)

    def __str__(self) -> str:
        return self.message


class ListOrderException(Exception):
    """
    Should be raised, if the ordering of a list is wrong (e.g. min values in an extent list after max...)

    :param msg: detailed error message
    """

    def __init__(self, msg: str) -> None:
        self.message = str(msg)

    def __str__(self) -> str:
        return self.message


class WellMarkerDepthException(Exception):
    """
    Should be raised, if a new well marker should be located deeper than the depth of the well

    :param msg: detailed error message
    """

    def __init__(self, msg: str) -> None:
        self.message = str(msg)

    def __str__(self) -> str:
        return self.message



# -*- coding: UTF-8 -*-
"""
This module provides a class for storing properties of points (and therefore also for lines). Properties for points are
the same as logs, except there is only one value, not a list of values. They are both derived from the
:class:`~geological_toolbox.abstract_log.AbstractLogClass`.
"""

import sqlalchemy as sq

from enum import Enum
from geological_toolbox.abstract_log import AbstractLogClass
from geological_toolbox.db_handler import Base, AbstractDBObject


class PropertyTypes(Enum):
    """Enum defining currently handled property types"""
    STRING = 0,
    """Representing a string"""
    INT = 1,
    """Representing an integer"""
    FLOAT = 2
    """Representing a float"""


class Property(Base, AbstractLogClass):
    """
    This class represents logging information for wells.
    For further details see :class:`~geological_toolbox.db_handler.AbstractLogClass`

    :param args: parameters for AbstractLogClass initialisation
    :param kwargs: parameters for AbstractLogClass initialisation
    """
    # define db table name and columns
    __tablename__ = "properties"

    id = sq.Column(sq.INTEGER, sq.Sequence("properties_id_seq"), primary_key=True)
    point_id = sq.Column(sq.INTEGER, sq.ForeignKey("geopoints.id"), default=-1)
    prop_type = sq.Column(sq.VARCHAR(20), default="STRING")
    prop_value = sq.Column(sq.TEXT, default="")

    def __init__(self, value: any, _type: PropertyTypes, *args, **kwargs) -> None:
        """
        Initialise the class
        """
        AbstractLogClass.__init__(self, *args, **kwargs)

        self.property_type = _type
        self.property_value = value

    def __repr__(self) -> str:
        text = "<Property(value={})>\n".format(self.prop_value)
        text += AbstractLogClass.__repr__(self)
        return text

    def __str__(self) -> str:
        text = "{} [{}]: {} - ".format(self.property_name, self.property_unit, self.prop_value)
        text += AbstractDBObject.__str__(self)
        return text

    def __convert_value(self, value: str) -> any or None:
        """
        converts the property value from type string to the specified type
        :return: converted property value
        """

        if self.property_type == PropertyTypes.STRING:
            return str(value)
        try:
            if self.property_type == PropertyTypes.INT:
                return int(value)
            if self.property_type == PropertyTypes.FLOAT:
                return float(value)
        except ValueError:
            return None

    def __check_value(self, value: any) -> bool:
        """
        Test, if the value can be converted to the specified format

        :param value: value to test
        :return: True, if it can be converted, else False
        """
        return self.__convert_value(value) is not None

    @property
    def property_type(self) -> PropertyTypes:
        """
        type of the property value
        :return: type of the property value
        :raises ValueError: if type is not available in PropertyTypes
        """
        return PropertyTypes[self.prop_type]

    @property_type.setter
    def property_type(self, value: PropertyTypes) -> None:
        """
        see getter
        """
        if not isinstance(value, PropertyTypes):
            raise ValueError("{} is not in PropertyTypes".format(value))
        self.prop_type = value.name

    @property
    def property_value(self) -> any or None:
        """
        converted value of the property
        :return: converted value of the property
        :raises ValueError: if prop_value cannot be converted to the specified property_type
        """
        return self.__convert_value(self.prop_value)

    @property_value.setter
    def property_value(self, value: any) -> None:
        """
        see getter
        """
        if self.__check_value(str(value)):
            self.prop_value = str(value)
        else:
            raise ValueError("Cannot convert property values [{}] to specified type {}".
                             format(value, self.property_type.name))



# -*- coding: UTF-8 -*-
"""
Module providing basic geometries (points and lines) for storing geological data in a database.
"""

import sqlalchemy as sq
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.orm import relationship
from sqlalchemy.orm.session import Session
from typing import List

from geological_toolbox.db_handler import Base, AbstractDBObject
from geological_toolbox.geo_object import AbstractGeoObject
from geological_toolbox.properties import Property
from geological_toolbox.stratigraphy import StratigraphicObject
from geological_toolbox.constants import float_precision


class GeoPoint(Base, AbstractGeoObject):
    """
    Represents a geological point feature, e.g. for storing line nodes or geological outcrops in a database.

    :param horizon: stratigraphic object of the point feature
    :param has_z: Is altitude stored as z value?
    :param args: parameters for AbstractGeoObject initialisation
    :param kwargs: parameters for AbstractGeoObject initialisation
    :return: Nothing
    :raises ValueError: if has_z is not compatible to type bool
    """
    # define db table name and columns
    __tablename__ = "geopoints"

    id = sq.Column(sq.INTEGER, sq.Sequence("geopoints_id_seq"), primary_key=True)
    has_z = sq.Column(sq.BOOLEAN, default=False)
    horizon_id = sq.Column(sq.INTEGER, sq.ForeignKey("stratigraphy.id"), default=-1)

    # define relationship to stratigraphic table
    hor = relationship("StratigraphicObject")

    line_id = sq.Column(sq.INTEGER, sq.ForeignKey("lines.id"), default=None, nullable=True)
    line_pos = sq.Column(sq.INTEGER, default=-1)

    # make the points unique -> coordinates + horizon + belongs to one line?
    sq.UniqueConstraint(AbstractGeoObject.east, AbstractGeoObject.north, AbstractGeoObject.alt, horizon_id,
                        AbstractGeoObject.name_col, line_id, line_pos, name="u_point_constraint")
    sq.Index("geopoint_coordinate_index", AbstractGeoObject.east, AbstractGeoObject.north)

    # add Property relation
    # order by property name
    properties = relationship("Property", order_by=Property.prop_name, backref="point",
                              primaryjoin="GeoPoint.id==Property.point_id",
                              cascade="all, delete, delete-orphan")

    def __init__(self, horizon: StratigraphicObject or None, has_z: bool, *args, **kwargs) -> None:
        """
        Creates a new GeoPoint
        """
        self.has_z = bool(has_z)
        self.horizon = horizon

        # call base class constructor
        AbstractGeoObject.__init__(self, *args, **kwargs)

    def __repr__(self) -> str:
        text = "<GeoPoint(id='{}', horizon='{}', line={}, line-position={})>\n". \
            format(self.id, str(self.horizon), self.line_id, self.line_pos)
        text += AbstractGeoObject.__repr__(self)
        text += "Properties: {}".format(self.properties)
        return text

    def __str__(self) -> str:
        text = "[{}] {} - {} - {}\n".format(self.id, str(self.horizon), self.line_id, self.line_pos)
        text += "AbstractGeoObject: {}\n".format(AbstractGeoObject.__str__(self))
        text += "Properties"
        # noinspection PyTypeChecker
        if len(self.properties) == 0:
            text += "\n  --- None ---"
        else:
            # noinspection PyTypeChecker
            for prop in self.properties:
                text += "\n  " + str(prop)

        return text

    @property
    def horizon(self) -> StratigraphicObject:
        """
        stratigraphic object of the point feature

        :raises TypeError: if value is not of type StratigraphicObject or None
        """
        return self.hor

    @horizon.setter
    def horizon(self, value: StratigraphicObject or None) -> None:
        """
        see getter
        """
        if (value is not None) and (type(value) is not StratigraphicObject):
            raise TypeError("type of committed value ({}) is not StratigraphicObject!".format(type(value)))

        if value is None:
            self.hor = None
            self.horizon_id = -1
        else:
            self.hor = value

    # delete z-dimension and information from point
    def del_z(self) -> None:
        """
        removes the z-value from the point, :attr:`~GeoObject.AbstractGeoObject.altitude` value will be set to 0

        :return: Nothing
        """
        self.alt = 0
        self.has_z = False

    def use_z(self) -> None:
        """
        uses the altitude value as z coordinate. Attention: please set the :attr:`ÃbstractGeoObject.altitude` property,
        if not it defaults to 0.

        :return: Nothing
        """
        self.has_z = True

    # add and remove property information form point
    def add_property(self, prop: Property) -> None:
        """
        Adds a new property to the point

        :param prop: new point property
        :return: Nothing
        :raises TypeError: if log is not of type Property
        """
        if type(prop) is not Property:
            raise TypeError("property {} is not of type Property!".format(str(prop)))

        self.properties.append(prop)

    def delete_property(self, prop: Property) -> None:
        """
        Deletes a property from the point

        :param prop: property to delete
        :return: Nothing
        :raises TypeError: if log is not of type Property
        :raises ValueError: if log is not part of self.properties
        """
        if type(prop) is not Property:
            raise TypeError("property {} is not of type Property!".format(str(prop)))

        try:
            self.properties.remove(prop)
        except ValueError as e:
            raise ValueError(str(e) + "\nProperty with ID " + str(prop.id) + " not found in list!")

    def has_property(self, property_name: str) -> bool:
        """
        Returns True, if a property with the name property_name exists for the GeoPoint, else False

        :param property_name: name of the requested property
        :return: True, if a property with the name property_name exists for the GeoPoint, else False
        """
        # noinspection PyTypeChecker
        for prop in self.properties:
            if prop.property_name == property_name:
                return True
        return False

    def get_property(self, property_name: str) -> Property:
        """
        Returns requested property, if the property with the name property_name exists for the GeoPoint

        :param property_name: name of the requested property
        :return: Returns the property
        :raises ValueError: if no property with the name exists
        """
        property_name = str(property_name)
        # noinspection PyTypeChecker
        for prop in self.properties:
            if prop.property_name == property_name:
                return prop
        raise ValueError("No property with the name '{}' exists!".format(property_name))

    # overwrite loading method
    @classmethod
    def load_all_without_lines_from_db(cls, session: Session) -> List["GeoPoint"]:
        """
        Returns all points in the database connected to the SQLAlchemy Session session, which are not part of a line.
        This function is similar the
        :meth:`AbstractDBObject.load_all_from_db()<geological_toolbox.db_handler.AbstractDBObject.load_all_from_db>`
        function.

        :param session: represents the database connection as SQLAlchemy Session
        :return: a list of points representing the result of the database query
        :raises TypeError: if session is not of type SQLAlchemy Session
        """
        if not isinstance(session, Session):
            raise TypeError("'session' is not of type SQLAlchemy Session!")

        # filter((GeoPoint.line_id is None) or (GeoPoint.line_id == -1) or (GeoPoint.line_id == "")). \
        result = session.query(GeoPoint). \
            filter(sq.or_(GeoPoint.line_id == None, GeoPoint.line_id == -1)). \
            order_by(cls.id).all()
        for point in result:
            point.session = session
        return result

    @classmethod
    def load_by_name_without_lines_from_db(cls, name: str, session: Session) -> List["GeoPoint"]:
        """
        Returns all GeoPoints with the given name in the database connected to the SQLAlchemy Session session, which
        are not part of a line. This function is similar the
        :meth:`AbstractDBObject.load_by_name_from_db()<geological_toolbox.db_handler.AbstractDBObject.load_by_name_from_db>`
        function.

        :param name: Only GeoPoints with this name will be returned
        :param session: represents the database connection as SQLAlchemy Session
        :return: a list of GeoPoints representing the result of the database query
        :raises TypeError: if session is not of type SQLAlchemy Session
        """
        if not isinstance(session, Session):
            raise TypeError("'session' is not of type SQLAlchemy Session!")

        result = session.query(GeoPoint). \
            filter(sq.or_(GeoPoint.line_id == None, GeoPoint.line_id == -1)). \
            filter(cls.name_col == name). \
            order_by(cls.id).all()
        for obj in result:
            obj.session = session
        return result

    @classmethod
    def load_in_extent_without_lines_from_db(cls, session: Session, min_easting: float, max_easting: float,
                                             min_northing: float, max_northing: float) -> List["GeoPoint"]:
        """
        Returns all points inside the given extent in the database connected to the SQLAlchemy Session session, which
        are not part of a line. This function is similar the
        :meth:`AbstractGeoObject.load_in_extent_from_db()<geological_toolbox.geo_object.AbstractGeoObject.load_in_extent_from_db>`
        function.

        :param min_easting: minimal easting of extent
        :param max_easting: maximal easting of extent
        :param min_northing: minimal northing of extent
        :param max_northing: maximal northing of extent
        :param session: represents the database connection as SQLAlchemy Session
        :return: a list of points representing the result of the database query
        :raises ValueError: if one of the extension values is not compatible to type float
        :raises TypeError: if session is not of type SQLAlchemy Session
        """
        min_easting = float(min_easting)
        max_easting = float(max_easting)
        min_northing = float(min_northing)
        max_northing = float(max_northing)

        if not isinstance(session, Session):
            raise TypeError("'session' is not of type SQLAlchemy Session!")

        result = session.query(GeoPoint).filter(sq.between(GeoPoint.east, min_easting, max_easting)). \
            filter(sq.between(GeoPoint.north, min_northing, max_northing)). \
            filter(sq.or_(GeoPoint.line_id == None, GeoPoint.line_id == -1)). \
            order_by(cls.id).all()
        for point in result:
            point.session = session
        return result


class Line(Base, AbstractDBObject):
    """
    Class Line

    Represents a geological line feature, e.g. for storing faults or horizon surface expressions.

    :param closed: True if the line should be closed, else False
    :param horizon: StratigraphicObject
    :param points: list of points which represents the lines nodes
    :param args: parameters for AbstractDBObject initialisation
    :param kwargs: parameters for AbstractDBObject initialisation
    :return: Nothing
    :raises ValueError: Raises ValueError if one of the types cannot be converted
    """

    # define db table name and columns
    __tablename__ = "lines"

    id = sq.Column(sq.INTEGER, sq.Sequence("lines_id_seq"), primary_key=True)
    closed = sq.Column(sq.BOOLEAN)

    # set stratigraphy of the line
    horizon_id = sq.Column(sq.INTEGER, sq.ForeignKey("stratigraphy.id"))
    hor = relationship("StratigraphicObject")

    # add GeoPoint relation, important is the ordering by line_pos value
    # collection_class function automatically reorders this value in case of insertion or deletion of points
    points = relationship("GeoPoint", order_by=GeoPoint.line_pos, collection_class=ordering_list("line_pos"),
                          backref="line", primaryjoin="Line.id==GeoPoint.line_id",
                          cascade="all, delete, delete-orphan")

    def __init__(self, closed: bool, horizon: StratigraphicObject, points: List[GeoPoint], *args, **kwargs) -> None:
        """
        Create a new line.
        """
        for pnt in points:
            if type(pnt) is not GeoPoint:
                raise ValueError("At least on point in points is not of type GeoPoint!")

        self.is_closed = bool(closed)
        self.horizon = horizon
        self.points = points

        # check doubled values and stratigraphy in the line
        self.__remove_doubled_points()
        self.__check_line()

        # call base class constructor
        AbstractDBObject.__init__(self, *args, **kwargs)

        if self.name_col == "":
            self.name_col = "line_{}".format(self.id)

    def __repr__(self) -> str:
        text = "<Line(id='{}', closed='{}', horizon='{}'\n)>".format(self.id, self.closed, str(self.horizon))
        text += AbstractDBObject.__repr__(self)
        text += "\npoints=\n" + str(self.points)
        return text

    def __str__(self) -> str:
        text = "[{}] {} - {}\n".format(self.id, "closed" if self.closed else "not closed", str(self.horizon))
        text += AbstractDBObject.__str__(self)
        text += "\npoints:\n"

        for point in self.points:
            text += "\n{}".format(str(point))

        return text

    @property
    def is_closed(self) -> bool:
        """
        True if the line is closed else False

        :raises ValueError: if value is not compatible to type bool
        """
        return self.closed

    @is_closed.setter
    def is_closed(self, value: bool) -> None:
        """
        see getter
        """
        value = bool(value)
        self.closed = bool(value)

    @property
    def horizon(self) -> StratigraphicObject:
        """
        The stratigraphy of the line object

        :raises TypeError: Raises TypeError if value is not of type Stratigraphy
        """
        return self.hor

    @horizon.setter
    def horizon(self, value: StratigraphicObject) -> None:
        """
        see getter
        """
        if (value is not None) and (type(value) is not StratigraphicObject):
            raise TypeError("type of committed value ({}) is not StratigraphicObject!".format(type(value)))

        if value is None:
            self.hor = None
            self.horizon_id = -1
        else:
            self.hor = value

        # check stratigraphy
        self.__check_line()

    def insert_point(self, point: GeoPoint, position: int) -> None:
        """
        Insert a point at the committed index.

        :param point: point to be inserted in the line
        :param position: Index, where points should be inserted in the line (starting with 0)
        :return: Nothing
        :raises TypeError: Raises TypeError if point is not of type GeoPoint
        :raises ValueError: Raises ValueError if position is not of type int or cannot be converted to int
        """
        position = int(position)
        if type(point) is not GeoPoint:
            raise TypeError("point {} is not of type GeoPoint!".format(str(point)))
        else:
            self.points.insert(position, point)

        # check doubled values and stratigraphy in the line
        self.__remove_doubled_points()
        self.__check_line()

    def insert_points(self, points: List[GeoPoint], position: int) -> None:
        """
        Insert a list of points at the committed index.

        :param points: List of points
        :param position: Index, where points should be inserted
        :return: Nothing
        :raises TypeError: Raises TypeError if one of the points is not of type GeoPoint
        :raises ValueError: Raises ValueError if position is not of type int or cannot be converted to int
        """
        position = int(position)

        for pnt in points:
            if type(pnt) is not GeoPoint:
                raise TypeError("At least on point in points is not of type GeoPoint!")

        # use slicing for points insert
        if len(self.points) <= position:
            self.points[-1:-1] = points
        else:
            self.points[position:position] = points

        # check doubled values and stratigraphy in the line
        self.__remove_doubled_points()
        self.__check_line()

    def get_point_index(self, point: GeoPoint) -> int:
        """
        Returns the index of the given point in the line

        :param point: point which has to be looked up
        :return: Index in the line
        :raises ValueError: Raises ValueError if committed point is not part of the line
        """
        return self.points.index(point)

    def delete_point(self, point: GeoPoint) -> None:
        """
        Deletes the point from the line

        :param point: GeoPoint object which should be deleted
        :return: Nothing
        :raises TypeError: Raises TypeError if point is not of type GeoPoint
        :raises ValueError: Raises ValueError the point is not part of the line
        """
        if type(point) is not GeoPoint:
            raise TypeError("point {} is not of type GeoPoint!".format(str(point)))

        try:
            self.points.remove(point)
        except ValueError as e:
            raise ValueError(str(e) + "\nGeoPoint with ID " + str(point.id) + " not found in list!")

        # check doubled values and stratigraphy in the line
        self.__remove_doubled_points()
        self.__check_line()

    def delete_point_by_coordinates(self, easting: float, northing: float, altitude: float = 0) -> None:
        """
        Deletes a point with the given coordinates from the line

        :param easting: easting value of the point to be deleted
        :param northing: northing value of the point to be deleted
        :param altitude: altitude value of the point to be deleted (only necessary if point has z-values!)
        :return: Nothing
        :raises ValueError: Raises ValueError if one parameter is not compatible to type float or no point can be found
        """
        try:
            easting = float(easting)
            northing = float(northing)
            altitude = float(altitude)
        except ValueError:
            raise ValueError("On of the input coordinates is not of type float!")

        # iterate over a copy of the list to avoid iteration issues caused by the deletion of values
        for pnt in self.points[:]:
            if (abs(float(pnt.easting) - easting) < float_precision) and (
                    abs(float(pnt.northing) - northing) < float_precision) and (
                    (abs(float(pnt.altitude) - altitude) < float_precision) or not pnt.has_z):
                self.points.remove(pnt)
                # check doubled values in a line
                self.__remove_doubled_points()
                self.__check_line()
                return

        raise ValueError("Point not found with coordinates {0}/{1}/{2}".format(easting, northing, altitude))

    def __check_line(self) -> None:
        """
        Checks the line for inconsistencies.
        Current checks:
        - stratigraphy check for all points

        :return: Nothing
        """
        for point in self.points:
            point.horizon = self.horizon

    def __remove_doubled_points(self) -> None:
        """
        remove points with the same coordinates, if they follow up each other in the points list.

        :return: None
        """
        # create a full copy of the point-list
        tmp_pnts = self.points[:]
        for i in range(len(tmp_pnts) - 1):
            coord_1 = (tmp_pnts[i].easting, tmp_pnts[i].northing, tmp_pnts[i].altitude)
            coord_2 = (tmp_pnts[i + 1].easting, tmp_pnts[i + 1].northing, tmp_pnts[i + 1].altitude)

            # remove doubled points
            if coord_1 == coord_2:
                self.points.remove(tmp_pnts[i + 1])

        # test for first and last point
        # if identical, then remove last point and set closed to True
        coord_1 = (self.points[0].easting, self.points[0].northing, self.points[0].altitude)
        coord_2 = (self.points[-1].easting, self.points[-1].northing, self.points[-1].altitude)

        if coord_1 == coord_2:
            del self.points[-1]
            self.closed = True

        # reorder to ensure right position value without reloading from database
        self.points.reorder()

    @classmethod
    def load_in_extent_from_db(cls, session: Session, min_easting: float, max_easting: float, min_northing: float,
                               max_northing: float) -> List["Line"]:
        """
        Returns all lines with at least on point inside the given extent in the database connected to the SQLAlchemy
        Session session. Overloads the
        :meth:`AbstractGeoObject.load_in_extent_from_db()<geological_toolbox.geo_object.AbstractGeoObject.load_in_extent_from_db>`
        function.

        :param min_easting: minimal easting of extent
        :param max_easting: maximal easting of extent
        :param min_northing: minimal northing of extent
        :param max_northing: maximal northing of extent
        :param session: represents the database connection as SQLAlchemy Session
        :return: a list of lines representing the result of the database query
        :raises ValueError: if one of the extension values is not compatible to type float
        :raises TypeError: if session is not of type SQLAlchemy Session
        """
        min_easting = float(min_easting)
        max_easting = float(max_easting)
        min_northing = float(min_northing)
        max_northing = float(max_northing)

        if not isinstance(session, Session):
            raise TypeError("'session' is not of type SQLAlchemy Session!")

        # select the points with a line_id that are located inside the extent
        # -> result will be a list of tuples with only a single line-id value
        # -> convert this list to a set in order to remove doubled values

        points = set(
            [x[0] for x in session.query(GeoPoint.line_id).
                filter(GeoPoint.line_id != -1).
                filter(sq.between(GeoPoint.east, min_easting, max_easting)).
                filter(sq.between(GeoPoint.north, min_northing, max_northing)).all()
             ])

        # return line with points in extend
        # to speed up the process, test if points (len > 0) exist in extent first
        if len(points) == 0:
            return []
        result = session.query(Line).filter(Line.id.in_(points)).order_by(cls.id).all()
        for line in result:
            line.session = session
        return result



"""
This module hosts the class Requests, which provides functionality for special (geo-)database requests.
"""

import sqlalchemy as sq
from sqlalchemy.orm.session import Session
from typing import List, Tuple

from geological_toolbox.exceptions import DatabaseException, DatabaseRequestException, FaultException, \
    ListOrderException
from geological_toolbox.geometries import GeoPoint
from geological_toolbox.properties import Property, PropertyTypes
from geological_toolbox.stratigraphy import StratigraphicObject
from geological_toolbox.wells import WellMarker


class Requests:
    """
    The class Requests, which provides functionality for special (geo-)database requests.
    """

    def __init__(self):
        # type: () -> None
        """
        Initialise the class

        -> Currently nothing to do...
        """
        pass

    @staticmethod
    def check_extent(extent: List[float]) -> None:
        """
        checks, if the given extent has the right format

        :param extent: value to be checked
        :type extent:
        :return: Nothing
        :raises TypeError: if extent is not a list
        :raises ValueError: if on list element is not compatible to float or number of elements is not 4
        :raises ListOrderException: if the ordering of the extent list [min_easting, max_easting, min_northing,
                max_northing] is wrong.
        """
        if extent is None:
            return

        if not isinstance(extent, List):
            raise TypeError("extent is not an instance of List()!")
        if len(extent) != 4:
            raise ValueError("Number of extension list elements is not 4!")
        for i in range(len(extent)):
            try:
                extent[i] = float(extent[i])
            except ValueError as e:
                raise ValueError("At least on extent element cannot be converted to float!\n{}".format(e))
        if extent[0] > extent[1]:
            raise ListOrderException("min easting > max easting")
        if extent[2] > extent[3]:
            raise ListOrderException("min northing > max northing")

    @staticmethod
    def create_thickness_point(sorted_dict: dict, well_id: int, marker_1: int, marker_2: int, session: Session,
                               use_faulted: bool = False, fault_name: str = "",
                               add_properties: Tuple = tuple()) -> GeoPoint:
        """
        Generate a new GeoPoint with thickness property from 2 well marker

        :param sorted_dict: dictionary containing well_id / WellMarker data
        :param well_id: current well_id
        :param marker_1: id of marker 1
        :param marker_2: id of marker 2
        :param session: current SQLAlchemy session
        :param use_faulted: should faulted sequence be included?
        :param fault_name: name of fault stratigraphic unit (default: "Fault")
        :param add_properties: Adds the properties to the GeoPoint. Format for each property: (value, type, name, unit)
        :return: new GeoPoint Object
        :raises FaultException: if a fault is inside the section and use_faulted is False
        :raises ValueError: if a property in the add_property tuple has less than 3 entries
        """

        min_depth = sorted_dict[well_id][marker_1].depth
        max_depth = sorted_dict[well_id][marker_2].depth

        faults = session.query(WellMarker).join(StratigraphicObject). \
            filter(WellMarker.horizon_id == StratigraphicObject.id). \
            filter(StratigraphicObject.unit_name == fault_name). \
            filter(WellMarker.well_id == well_id)
        if min_depth > max_depth:
            faults = faults.filter(sq.between(WellMarker.drill_depth, max_depth, min_depth))
        else:
            faults = faults.filter(sq.between(WellMarker.drill_depth, min_depth, max_depth))
        if (faults.count() > 0) and (use_faulted is False):
            raise FaultException("Fault inside section")

        point = sorted_dict[well_id][marker_1].to_geopoint()
        thickness = Property(max_depth - min_depth, PropertyTypes.FLOAT, "thickness", "m", session)
        point.add_property(thickness)
        if use_faulted:
            faulted = Property(faults.count(), PropertyTypes.INT, "faulted", "count", session)
            point.add_property(faulted)
        for prop in add_properties:
            if len(prop) < 4:
                raise ValueError("property tuple has less than 4 entries!")
            new_property = Property(prop[0], PropertyTypes[prop[1]], prop[2], prop[3], session)
            point.add_property(new_property)
        return point

    @staticmethod
    def well_markers_to_thickness(session: Session, marker_1: str, marker_2: str, summarise_multiple: bool = False,
                                  extent: Tuple[int, int, int, int] or None = None, use_faulted: bool = False,
                                  fault_name: str = "Fault") -> List[GeoPoint]:
        """
        Static method for generating a point set including a thickness property derived from the committed well marker

        :param session: The SQLAlchemy session connected to the database storing the geodata
        :param marker_1: First WellMarker unit name
        :param marker_2: Second WellMarker unit name
        :param summarise_multiple: Summarise multiple occurrences of marker_1 and marker_2 to a maximum thickness.
               If this parameter is False (default value) create multiple points.
        :param extent: extension rectangle as list which borders the well distribution. The list has the following
               order: (min easting, max easting, min northing, max northing)
        :param use_faulted: if True, also sections with faults between marker_1 and marker_2 are returned
        :param fault_name: unit name of fault marker (default: "Fault")
        :return: A list of GeoPoints each with a thickness property
        :raises AttributeError: if marker_1 and marker_2 are equal
        :raises DatabaseException: if the database query results in less than 2 marker of a well_id
        :raises DatabaseRequestException: if an unexpected query result occurs
        :raises TypeError: if session is not an instance of SQLAlchemy session
        :raises ValueError: if a parameter is not compatible with the required type

        for further raises see :meth:`Requests.check_extent`

        Query for selecting markers:

        .. code-block:: sql
            :linenos:

            SELECT wm1.* FROM well_marker wm1
            JOIN stratigraphy st1
            ON wm1.horizon_id = st1.id
            WHERE st1.unit_name IN ("mu", "so")
            AND EXISTS
            (
                SELECT 1 FROM well_marker wm2
                JOIN stratigraphy st2
                ON wm2.horizon_id = st2.id
                WHERE st2.unit_name IN ("mu", "so")
                AND wm1.well_id = wm2.well_id
                AND st1.unit_name <> st2.unit_name
            )
            ORDER BY wm1.well_id,wm1.drill_depth
        """
        if not isinstance(session, Session):
            raise TypeError("session is not of type SQLAlchemy Session")

        if marker_1 == marker_2:
            raise AttributeError("marker_1 and marker_2 cannot be equal!")

        if extent is not None:
            Requests.check_extent(extent)

        result = session.query(WellMarker)
        if extent is None:
            statement = sq.text("SELECT wm1.* FROM well_marker wm1 " +
                                "JOIN stratigraphy st1 ON wm1.horizon_id = st1.id " +
                                "WHERE st1.unit_name IN (:marker1, :marker2) " +
                                "AND EXISTS " +
                                "( SELECT 1 FROM well_marker wm2 JOIN stratigraphy st2 ON wm2.horizon_id = st2.id " +
                                "WHERE st2.unit_name IN (:marker1, :marker2) AND wm1.well_id = wm2.well_id " +
                                "AND st1.unit_name <> st2.unit_name) ORDER BY wm1.well_id,wm1.drill_depth")
            result = result.from_statement(statement). \
                params(marker1=marker_1, marker2=marker_2). \
                all()
        else:
            statement = sq.text("SELECT wm1.* FROM well_marker wm1 " +
                                "JOIN wells ON wm1.well_id = wells.id " +
                                "JOIN stratigraphy st1 ON wm1.horizon_id = st1.id " +
                                "WHERE wells.east BETWEEN :east_min AND :east_max " +
                                "AND wells.north BETWEEN :north_min AND :north_max " +
                                "AND st1.unit_name IN (:marker1, :marker2)" +
                                "AND EXISTS " +
                                "( SELECT 1 FROM well_marker wm2 JOIN stratigraphy st2 ON wm2.horizon_id = st2.id " +
                                "WHERE st2.unit_name IN (:marker1, :marker2) AND wm1.well_id = wm2.well_id " +
                                "AND st1.unit_name <> st2.unit_name) ORDER BY wm1.well_id,wm1.drill_depth")
            result = result.from_statement(statement). \
                params(marker1=marker_1, marker2=marker_2, east_min=extent[0], east_max=extent[1], north_min=extent[2],
                       north_max=extent[3]). \
                all()

        # first: sort by well_id for simpler multiple marker check
        sorted_dict = dict()
        for i in range(len(result)):
            if result[i].well_id not in sorted_dict:
                sorted_dict[result[i].well_id] = list()
            sorted_dict[result[i].well_id].append(result[i])

        del result

        # generate the resulting list of GeoPoints
        geopoints = list()
        for well_id in sorted_dict:
            if len(sorted_dict[well_id]) < 2:
                raise DatabaseException("Not enough well marker in dictionary")
            if len(sorted_dict[well_id]) == 2:
                sorted_dict[well_id][0].session = session
                try:
                    if summarise_multiple:
                        point = Requests.create_thickness_point(
                            sorted_dict, well_id, 0, 1, session, use_faulted, fault_name,
                            ((0, "INT", "summarised", "bool"),))
                        geopoints.append(point)
                    else:
                        point = Requests.create_thickness_point(
                            sorted_dict, well_id, 0, 1, session, use_faulted, fault_name,
                            ((0, "INT", "multiple marker", "bool"),))
                        geopoints.append(point)
                # FaultException -> do nothing except catching the exception
                except FaultException:
                    continue
                # don't test anything else for this well_id
                continue

            # last case: more than 2 values found:
            if summarise_multiple:
                first_index = -1
                last_index = -1
                for i in range(len(sorted_dict[well_id])):
                    if sorted_dict[well_id][i].horizon.unit_name == marker_1:
                        first_index = i
                        break
                for j in reversed(range(len(sorted_dict[well_id]))):
                    if sorted_dict[well_id][j].horizon.unit_name == marker_2:
                        last_index = j
                        break

                if (first_index == -1) or (last_index == -1):
                    raise DatabaseRequestException("Didn't find two different markers. Shouldn't be possible. " +
                                                   "Please excuse this error and forward it to me.")

                # wrong order -> nothing to do...
                if last_index < first_index:
                    continue

                try:
                    sorted_dict[well_id][first_index].session = session
                    point = Requests.create_thickness_point(
                        sorted_dict, well_id, first_index, last_index, session, use_faulted, fault_name,
                        ((1, "INT", "summarised", "bool"),))
                    geopoints.append(point)
                # FaultException -> do nothing except catching the exception
                except FaultException:
                    continue

                # finished summarise section -> continue to avoid second round without summarise
                continue
            # don't summarise
            first_index = -1
            for index in range(len(sorted_dict[well_id])):
                if sorted_dict[well_id][index].horizon.unit_name == marker_1:
                    first_index = index
                elif (first_index != -1) and (sorted_dict[well_id][index].horizon.unit_name == marker_2):
                    try:
                        sorted_dict[well_id][first_index].session = session
                        point = Requests.create_thickness_point(
                            sorted_dict, well_id, first_index, index, session, use_faulted, fault_name,
                            ((1, "INT", "multiple marker", "bool"),))
                        geopoints.append(point)
                    # FaultException -> do nothing except catching the exception
                    except FaultException:
                        continue
                    finally:
                        first_index = -1

        return geopoints

    @staticmethod
    def interpolate_geopoints(points: GeoPoint, property_name: str, method: str) -> None:
        """
        Interpolate the property values of the given GeoPoints using the interpolation method "method"

        .. todo:: - Integrate functionality
                  - define interpolation methods
                  - define return statement

        :param points: List of GeoPoints as interpolation base
        :param property_name: Name of the property to interpolate
        :param method: Interpolation method
        :return: Currently Nothing
        :raises TypeError: if on of the points is not of type GeoPoint

        possible values for interpolation **method** are:

        - **nearest** (Nearest Neighbour interpolation)
        - **ide** (Inverse Distance interpolation)
        - **spline** (Thin-Plate-Spline interpolation)
        """



# -*- coding: UTF-8 -*-
"""
Summarized import of database objects to ensure correct alembic migration script generation
"""

from geological_toolbox.geometries import GeoPoint, Line
from geological_toolbox.properties import Property
from geological_toolbox.stratigraphy import StratigraphicObject
from geological_toolbox.wells import Well, WellMarker
from geological_toolbox.well_logs import WellLog, WellLogValue
from geological_toolbox.db_handler import Base



# -*- coding: UTF-8 -*-
"""
Definition of package wide constants.
"""

from ._version import __version__

project_version = __version__.split('.')
"""
Current project version
"""

float_precision = 0.001  # 1 mm
"""
precision for float comparison
"""



"""db_initialization

Revision ID: 2ec5e0e664cf
Revises: 
Create Date: 2019-09-27 14:24:20.961906

"""
from alembic import op
import sqlalchemy as sq


# revision identifiers, used by Alembic.
revision = '2ec5e0e664cf'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('stratigraphy',
    sq.Column('name_col', sq.VARCHAR(length=100), nullable=True),
    sq.Column('comment_col', sq.VARCHAR(length=100), nullable=True),
    sq.Column('id', sq.INTEGER(), nullable=False),
    sq.Column('unit_name', sq.VARCHAR(length=50), nullable=True),
    sq.Column('age', sq.FLOAT(), nullable=True),
    sq.PrimaryKeyConstraint('id'),
    sq.UniqueConstraint('unit_name')
    )
    op.create_table('wells',
    sq.Column('name_col', sq.VARCHAR(length=100), nullable=True),
    sq.Column('comment_col', sq.VARCHAR(length=100), nullable=True),
    sq.Column('east', sq.FLOAT(), nullable=True),
    sq.Column('north', sq.FLOAT(), nullable=True),
    sq.Column('alt', sq.FLOAT(), nullable=True),
    sq.Column('reference', sq.TEXT(), nullable=True),
    sq.Column('id', sq.INTEGER(), nullable=False),
    sq.Column('drill_depth', sq.FLOAT(), nullable=True),
    sq.Column('wellname', sq.VARCHAR(length=100), nullable=True),
    sq.Column('shortwellname', sq.VARCHAR(length=100), nullable=True),
    sq.PrimaryKeyConstraint('id'),
    sq.UniqueConstraint('wellname')
    )
    op.create_table('lines',
    sq.Column('name_col', sq.VARCHAR(length=100), nullable=True),
    sq.Column('comment_col', sq.VARCHAR(length=100), nullable=True),
    sq.Column('id', sq.INTEGER(), nullable=False),
    sq.Column('closed', sq.BOOLEAN(), nullable=True),
    sq.Column('horizon_id', sq.INTEGER(), nullable=True),
    sq.ForeignKeyConstraint(['horizon_id'], ['stratigraphy.id'], ),
    sq.PrimaryKeyConstraint('id')
    )
    op.create_table('well_logs',
    sq.Column('name_col', sq.VARCHAR(length=100), nullable=True),
    sq.Column('comment_col', sq.VARCHAR(length=100), nullable=True),
    sq.Column('prop_name', sq.VARCHAR(length=50), nullable=True),
    sq.Column('prop_unit', sq.VARCHAR(length=100), nullable=True),
    sq.Column('id', sq.INTEGER(), nullable=False),
    sq.Column('well_id', sq.INTEGER(), nullable=True),
    sq.ForeignKeyConstraint(['well_id'], ['wells.id'], ),
    sq.PrimaryKeyConstraint('id')
    )
    op.create_table('well_marker',
    sq.Column('name_col', sq.VARCHAR(length=100), nullable=True),
    sq.Column('comment_col', sq.VARCHAR(length=100), nullable=True),
    sq.Column('id', sq.INTEGER(), nullable=False),
    sq.Column('drill_depth', sq.FLOAT(), nullable=True),
    sq.Column('horizon_id', sq.INTEGER(), nullable=True),
    sq.Column('well_id', sq.INTEGER(), nullable=True),
    sq.ForeignKeyConstraint(['horizon_id'], ['stratigraphy.id'], ),
    sq.ForeignKeyConstraint(['well_id'], ['wells.id'], ),
    sq.PrimaryKeyConstraint('id')
    )
    op.create_table('geopoints',
    sq.Column('name_col', sq.VARCHAR(length=100), nullable=True),
    sq.Column('comment_col', sq.VARCHAR(length=100), nullable=True),
    sq.Column('east', sq.FLOAT(), nullable=True),
    sq.Column('north', sq.FLOAT(), nullable=True),
    sq.Column('alt', sq.FLOAT(), nullable=True),
    sq.Column('reference', sq.TEXT(), nullable=True),
    sq.Column('id', sq.INTEGER(), nullable=False),
    sq.Column('has_z', sq.BOOLEAN(), nullable=True),
    sq.Column('horizon_id', sq.INTEGER(), nullable=True),
    sq.Column('line_id', sq.INTEGER(), nullable=True),
    sq.Column('line_pos', sq.INTEGER(), nullable=True),
    sq.ForeignKeyConstraint(['horizon_id'], ['stratigraphy.id'], ),
    sq.ForeignKeyConstraint(['line_id'], ['lines.id'], ),
    sq.PrimaryKeyConstraint('id')
    )
    op.create_table('logging_association',
    sq.Column('name_col', sq.VARCHAR(length=100), nullable=True),
    sq.Column('comment_col', sq.VARCHAR(length=100), nullable=True),
    sq.Column('id', sq.INTEGER(), nullable=False),
    sq.Column('log_depth', sq.FLOAT(), nullable=True),
    sq.Column('log_value', sq.FLOAT(), nullable=True),
    sq.Column('log_id', sq.INTEGER(), nullable=True),
    sq.ForeignKeyConstraint(['log_id'], ['well_logs.id'], ),
    sq.PrimaryKeyConstraint('id')
    )
    op.create_index('welllogdepth_index', 'logging_association', ['log_depth'], unique=False)
    op.create_table('properties',
    sq.Column('name_col', sq.VARCHAR(length=100), nullable=True),
    sq.Column('comment_col', sq.VARCHAR(length=100), nullable=True),
    sq.Column('prop_name', sq.VARCHAR(length=50), nullable=True),
    sq.Column('prop_unit', sq.VARCHAR(length=100), nullable=True),
    sq.Column('id', sq.INTEGER(), nullable=False),
    sq.Column('point_id', sq.INTEGER(), nullable=True),
    sq.Column('prop_type', sq.VARCHAR(length=20), nullable=True),
    sq.Column('prop_value', sq.TEXT(), nullable=True),
    sq.ForeignKeyConstraint(['point_id'], ['geopoints.id'], ),
    sq.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('properties')
    op.drop_index('welllogdepth_index', table_name='logging_association')
    op.drop_table('logging_association')
    op.drop_table('geopoints')
    op.drop_table('well_marker')
    op.drop_table('well_logs')
    op.drop_table('lines')
    op.drop_table('wells')
    op.drop_table('stratigraphy')
    # ### end Alembic commands ###



# -*- coding: UTF-8 -*-
"""
The setuptools base setup module. Necessary for the PyPI upload.
"""

# Always prefer setuptools over distutils
from setuptools import setup
# load version
from geological_toolbox.constants import project_version
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))
status = "Development Status :: "
if project_version[3].startswith("a"):
    status += "3 - Alpha"
elif project_version[3].startswith("b"):
    status += "4 - Beta"
elif project_version[3].startswith("rc") or project_version[3].startswith(""):
    status += "5 - Production/Stable"
# elif project_version[1] == "0" and project_version[3].startswith("rc"):
#     status += "6 - Mature"
else:
    status += "1 - Planning"

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()
long_description_content_type = "text/markdown"

setup(
    # $ pip install GeologicalToolbox
    #
    # PyPI: https://pypi.org/project/GeologicalToolbox/

    name="GeologicalToolbox",
    version=".".join(project_version),
    description="A python module for processing and storing geological data.",
    long_description=long_description,  # Optional
    long_description_content_type=long_description_content_type,
    url="https://github.com/stdonn/GeologicalToolbox",  # Optional
    author="Stephan Donndorf",  # Optional
    author_email="stephan@donndorf.info",  # Optional
    license="MIT",

    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers

    classifiers=[
        status,
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Intended Audience :: Other Audience",
        "Intended Audience :: Science/Research",

        # Pick your license as you wish
        "License :: OSI Approved :: MIT License",

        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",

        # other classifiers
        "Natural Language :: English",
        "Topic :: Database :: Front-Ends",
        "Topic :: Documentation :: Sphinx",
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],

    keywords="Geology Modelling SQLAlchemy",  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=["geological_toolbox"],  # Required
    package_data={"geological_toolbox": ["alembic.ini", "alembic/env.py", "alembic/versions/*.py"]},

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip"s requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "SQLAlchemy>=1.2",
        "alembic>=1.2"
    ],  # Optional

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    extras_require={  # Optional
        ":python_version < '3.5'": [
            "typing"
        ],
        "dev": ["Sphinx", "sphinxjp.themes.basicstrap"]
        # "test": ["coverage"],
    },

    python_requires=">=3.5"
)



# -*- coding: UTF-8 -*-
"""
This module provides classes for storage of drilling data in a database.
"""

import sqlalchemy as sq
from sqlalchemy.orm import relationship
from sqlalchemy.orm.session import Session
from typing import List

from geological_toolbox.exceptions import DatabaseException, WellMarkerDepthException
from geological_toolbox.geo_object import AbstractGeoObject
from geological_toolbox.geometries import GeoPoint
from geological_toolbox.well_logs import WellLog
from geological_toolbox.db_handler import Base, AbstractDBObject
from geological_toolbox.stratigraphy import StratigraphicObject


class WellMarker(Base, AbstractDBObject):
    """
    Represents single markers in a drilled well

    :param depth: depth below Kelly Bushing
    :param horizon: stratigraphic object
    :raises TypeError: if one of the committed parameters cannot be converted to the expected type
    """
    __tablename__ = "well_marker"

    id = sq.Column(sq.INTEGER, sq.Sequence("well_marker_id_seq"), primary_key=True)
    drill_depth = sq.Column(sq.FLOAT)

    # define relationship to stratigraphic table
    horizon_id = sq.Column(sq.INTEGER, sq.ForeignKey("stratigraphy.id"))
    hor = relationship("StratigraphicObject")

    well_id = sq.Column(sq.INTEGER, sq.ForeignKey("wells.id"), default=-1)

    def __init__(self, depth: float, horizon: StratigraphicObject, *args, **kwargs) -> None:
        """
        initialize the class
        """
        AbstractDBObject.__init__(self, *args, **kwargs)

        self.depth = float(depth)
        self.horizon = horizon

    def __repr__(self) -> str:
        return "<WellMarker(id='{}', depth='{}', horizon='{}'\nAbstractObject: {}". \
            format(self.id, self.depth, self.horizon, AbstractDBObject.__repr__(self))

    def __str__(self) -> str:
        return "[{}] {}: {} - {}".format(self.id, self.depth, self.horizon, AbstractDBObject.__str__(self))

    @property
    def depth(self) -> float:
        """
        depth of the well marker

        :return: depth of the well marker.
        :raises ValueError: if value if not of type float or cannot be converted to float
        """
        return float(self.drill_depth)

    @depth.setter
    def depth(self, value: float) -> None:
        """
        see getter
        """
        self.drill_depth = float(value)

    @property
    def horizon(self) -> StratigraphicObject:
        """
        stratigraphic object of the point

        :return: stratigraphic object of the point
        :raises TypeError: if value is not of type StratigraphicObject
        """
        return self.hor

    @horizon.setter
    def horizon(self, value: StratigraphicObject) -> None:
        """
        see getter
        """
        if (value is not None) and (type(value) is not StratigraphicObject):
            raise TypeError("type of committed value ({}) is not StratigraphicObject!".format(type(value)))

        if value is None:
            self.hor = None
            self.horizon_id = -1
        else:
            self.hor = value

    def to_geopoint(self) -> GeoPoint:
        """
        Returns the current well marker as GeoPoint

        :return: the current well marker as GeoPoint
        """
        easting = self.well.easting
        northing = self.well.northing
        altitude = self.well.altitude - self.depth
        return GeoPoint(self.horizon, True, self.well.reference_system, easting, northing, altitude, self.session,
                        self.well.well_name, self.comment)

    @classmethod
    def load_in_extent_from_db(cls, session: Session, min_easting: float, max_easting: float, min_northing: float,
                               max_northing: float) -> List["WellMarker"]:
        """
        Returns all well marker within the given extent in the database connected to the SQLAlchemy Session session

        :param min_easting: minimal easting of extent
        :param max_easting: maximal easting of extent
        :param min_northing: minimal northing of extent
        :param max_northing: maximal northing of extent
        :param session: represents the database connection as SQLAlchemy Session
        :return: a list of WellMarker representing the result of the database query
        """
        result = session.query(cls, Well). \
            filter(Well.id == cls.well_id). \
            filter(sq.between(Well.east, min_easting, max_easting)). \
            filter(sq.between(Well.north, min_northing, max_northing)). \
            order_by(cls.id).all()
        for marker in result:
            marker.session = session
        return result

    @classmethod
    def load_all_by_stratigraphy_from_db(cls, horizon: StratigraphicObject, session: Session) -> List["WellMarker"]:
        """
        Returns all WellMarker in the database which are related to the StratigraphicObject "horizon"

        :param horizon: stratigraphic object for the database query
        :param session: represents the database connection as SQLAlchemy Session
        :return: a list of WellMarker
        """
        return session.query(cls).filter(WellMarker.horizon_id == horizon.id)

    @classmethod
    def load_all_by_stratigraphy_in_extent_from_db(cls, horizon: StratigraphicObject, min_easting: float,
                                                   max_easting: float, min_northing: float, max_northing: float,
                                                   session: Session) -> List["WellMarker"]:
        """
        Returns all WellMarker in the database which are related to the StratigraphicObject "horizon" and are located
        within the given extent

        :param horizon: stratigraphic object for the database query
        :param min_easting: minimal easting of extent
        :param max_easting: maximal easting of extent
        :param min_northing: minimal northing of extent
        :param max_northing: maximal northing of extent
        :param session: represents the database connection as SQLAlchemy Session
        :return: a list of WellMarker
        """
        result = session.query(cls, Well). \
            filter(cls.horizon_id == horizon.id). \
            filter(Well.id == cls.well_id). \
            filter(sq.between(Well.east, min_easting, max_easting)). \
            filter(sq.between(Well.north, min_northing, max_northing)). \
            order_by(cls.id).all()
        for marker in result:
            marker.session = session
        return result


class Well(Base, AbstractGeoObject):
    """
    Represents a well for storage of geoscientific data.

    :param session: SQLAlchemy session, which includes the database connection
    :param well_name: well name
    :param short_name: shortened well name
    :param depth: drilled depth of the well
    :raises ValueError: Raises ValueError if one of the types cannot be converted
    """
    __tablename__ = "wells"

    id = sq.Column(sq.INTEGER, sq.Sequence("wells_id_seq"), primary_key=True)
    drill_depth = sq.Column(sq.FLOAT)
    wellname = sq.Column(sq.VARCHAR(100), unique=True)
    shortwellname = sq.Column(sq.VARCHAR(100), default="")

    # define markers relationship
    marker = relationship("WellMarker", order_by=WellMarker.drill_depth,
                          backref="well", primaryjoin="Well.id == WellMarker.well_id",
                          cascade="all, delete, delete-orphan")

    logs = relationship("WellLog", order_by=WellLog.id,
                        backref="well", primaryjoin="Well.id == WellLog.well_id",
                        cascade="all, delete, delete-orphan")

    sq.Index("coordinate_index", AbstractGeoObject.east, AbstractGeoObject.north)

    def __init__(self, well_name: str, short_name: str, depth: float, *args, **kwargs) -> None:
        """
        initialize the class
        """

        self.depth = depth
        self.well_name = well_name
        self.short_name = short_name

        # call base class constructor
        AbstractGeoObject.__init__(self, *args, **kwargs)

    def __repr__(self) -> str:
        return "<Well(id='{}', well_name='{}', short_name='{}', depth='{}', marker='{}', logs='{}', comment='{}'," + \
               " marker='{}')>".format(self.id, self.well_name, self.short_name, self.easting, self.northing,
                                       self.altitude, self.depth, self.comment, repr(self.marker))

    def __str__(self) -> str:
        text = "[{}] {} ({}):\n{} - {} - {} - {} - {}" \
            .format(self.id, self.well_name, self.short_name, self.easting, self.northing, self.altitude, self.depth,
                    self.comment)

        for marker in self.marker:
            text += "\n" + str(marker)

        return text

    @property
    def depth(self) -> float:
        """
        total drilling depth of the well

        :raises ValueError: Raises ValueError if depth is not of type float, it cannot be converted to float or when
                            depth is < 0
        """
        return float(self.drill_depth)

    @depth.setter
    def depth(self, dep: float) -> None:
        """
        see getter
        """
        dep = float(dep)
        if dep < 0:
            raise ValueError("Depth is below 0! ({})".format(dep))
        if (len(self.marker) > 0) and (dep < self.marker[-1].depth):
            raise WellMarkerDepthException(
                "New depth ({}) lower than depth of last marker {}".format(dep, self.marker[-1].depth))
        self.drill_depth = dep

    @property
    def well_name(self) -> str:
        """
        unique name of the well
        """
        return self.wellname

    @well_name.setter
    def well_name(self, well_name: str) -> None:
        """
        see getter
        """
        well_name = str(well_name)
        if len(well_name) > 100:
            well_name = well_name[:100]
        self.wellname = well_name

    @property
    def short_name(self) -> str:
        """
        shortend name of the well
        """
        return self.shortwellname

    @short_name.setter
    def short_name(self, short_name: str) -> None:
        """
        see getter
        """
        short_name = str(short_name)
        if len(short_name) > 20:
            short_name = short_name[:20]
        self.shortwellname = short_name

    def insert_marker(self, marker: WellMarker) -> None:
        """
        Insert a new WellMarker
        ATTENTION: If you insert a marker, the well will be automatically stored in the database!

        :param marker: WellMarker to be inserted
        :return: Nothing
        :raises TypeError: if marker is not of type WellMarker
        :raises ValueError: if the depth of the marker is larger than the drilled depth of the well
        """
        if type(marker) is not WellMarker:
            raise TypeError("marker {} is not of type WellMarker!".format(str(marker)))
        if marker.depth > self.depth:
            raise ValueError("Marker depth ({}) is larger than final well depth ({})!".format(marker.depth, self.depth))
        self.marker.append(marker)

        # new sorting to ensure correct order without storage and reloading from the database
        self.marker.sort(key=lambda x: x.depth)

    def insert_multiple_marker(self, marker: List[WellMarker]) -> None:
        """
        Insert multiple marker in the well
        ATTENTION: If you insert marker, the well will be automatically stored in the database!

        :param marker: List of marker to be inserted
        :return: Nothing
        :raises TypeError: if one of the marker is not of type WellMarker
        :raises ValueError: if the depth of a marker is larger than the drilled depth of the well
        """
        for mark in marker:
            if type(mark) is not WellMarker:
                raise TypeError(
                    "At least on marker is not of type WellMarker ({}: {})!".format(str(mark), str(type(mark))))
            if mark.depth > self.depth:
                raise ValueError("Marker depth ({}) is larger than final well depth ({})!".
                                 format(mark.depth, self.depth))

        self.marker += marker

        # new sorting to ensure correct order without storage and reloading from the database
        self.marker.sort(key=lambda x: x.depth)

    def get_marker_by_depth(self, depth: float) -> WellMarker or None:
        """
        Returns the marker at depth "depth"

        :param depth: depth of the requested marker
        :return: the marker at depth "depth"
        :raises ValueError: if no marker was found for the committed depth or depth is not compatible to float
        """
        depth = float(depth)
        for marker in self.marker:
            if marker.depth == depth:
                return marker
        raise ValueError("No marker found at depth {}".format(depth))

    def delete_marker(self, marker: WellMarker) -> None:
        """
        Deletes the marker from the well

        :param marker: WellMarker object which should be deleted
        :return: Nothing
        :raises TypeError: if marker is not of type WellMarker
        :raises ValueError: the marker is not part of the well
        """
        if type(marker) is not WellMarker:
            raise TypeError("marker {} is not of type WellMarker!".format(str(marker)))

        try:
            self.marker.remove(marker)
            WellMarker.delete_from_db(marker, self.session)
        except ValueError as e:
            raise ValueError(str(e) + "\nWellMarker with ID " + str(marker.id) + " not found in list!")

    def add_log(self, log: WellLog) -> None:
        """
        Adds a new log to the well

        :param log: new well log
        :return: Nothing
        :raises TypeError: if log is not of type WellLog
        """
        if type(log) is not WellLog:
            raise TypeError("log {} is not of type WellLog!".format(str(log)))

        self.logs.append(log)

    def delete_log(self, log: WellLog) -> None:
        """
        Deletes the log from the well

        :param log: log to delete
        :return: Nothing
        :raises TypeError: if log is not of type WellLog
        :raises ValueError: if log is not part of the well
        """
        if type(log) is not WellLog:
            raise TypeError("log {} is not of type WellLog!".format(str(log)))

        try:
            self.logs.remove(log)
        except ValueError as e:
            raise ValueError(str(e) + "\nWellLog with ID " + str(log.id) + " not found in list!")

    @classmethod
    def load_by_wellname_from_db(cls, name: str, session: Session) -> "Well" or None:
        """
        Returns the well with the given name in the database connected to the SQLAlchemy Session session

        :param name: name of the requested well
        :param session: represents the database connection as SQLAlchemy Session
        :return: As the name is a unique value, only one result can be returned or None
        :raises DatabaseException: if more than one result was found (the well name is an unique value)
        """
        result = session.query(cls).filter(cls.wellname == name)
        if result.count() == 0:
            return None
        if result.count() == 1:
            result = result.one()
            result.session = session
            return result

        raise DatabaseException("More than one ({}) well with the same name: {}! Database error!".
                                format(result.count(), name))

    @classmethod
    def load_deeper_than_value_from_db(cls, session: Session, min_depth: float) -> List["Well"]:
        """
        Returns all wells with a drilled depth larger than min_depth in the database connected to the SQLAlchemy Session
        session

        :param min_depth: minimal drilled depth
        :param session: represents the database connection as SQLAlchemy Session
        :return: a list of wells representing the result of the database query
        """
        result = session.query(cls).filter(cls.drill_depth >= min_depth)
        result = result.order_by(cls.id).all()
        for well in result:
            well.session = session
        return result



# -*- coding: UTF-8 -*-
"""
Providing the DBHandler class for database access through a SQLAlchemy session and the AbstractDBObject base class for
all other database related classes.
"""

import inspect
import os
import sqlalchemy as sq

from alembic import command, script
from alembic.config import Config
from alembic.runtime import migration
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm.session import Session
from typing import List

from geological_toolbox.exceptions import DatabaseRequestException

Base = declarative_base()


class DBHandler(object):
    """
    A class for database access through an SQLAlchemy session.
    Same initialisation as SQLAlchemy provides. Additional arguments are piped to :meth:`sqlalchemy.create_engine`

    :param connection: Connection uri to a database, format defined by SQLAlchemy
    :return: Nothing
    """

    def __init__(self, connection: str = "sqlite:///:memory:", *args, **kwargs) -> None:
        """
        Initialize a new database connection via SQLAlchemy
        """
        self.__connection = connection
        self.__args = args
        self.__kwargs = kwargs

        self.__engine = sq.create_engine(self.__connection, *self.__args, **self.__kwargs)
        self.__last_session = None
        self.__config = "alembic.ini"

        in_memory = self.__connection in ["sqlite://", "sqlite:///:memory:"]
        if not (in_memory or self.check_current_head()):
            self.start_db_migration()

        if in_memory:
            Base.metadata.create_all(self.__engine)

        self.__sessionmaker = sessionmaker(bind=self.__engine)
        self.create_new_session()

    def check_current_head(self) -> bool:
        """
        Checks if the selected database schema version matches the python source ORM schema version.
        If false please run start_db_migration to update the database.

        :return: True if both version are matching, else False
        """
        local_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))  # script directory
        alembic_cfg = Config(os.path.abspath(os.path.join(local_dir, "alembic.ini")))

        # Alembic has problems, when called inside a library,
        # therefore we change the current folder with its absolute path
        alembic_cfg.set_main_option(
            "script_location",
            os.path.abspath(os.path.join(local_dir, alembic_cfg.get_main_option("script_location", "alembic")))
        )

        directory = script.ScriptDirectory.from_config(alembic_cfg)
        with self.__engine.begin() as connection:
            context = migration.MigrationContext.configure(connection)
            return set(context.get_current_heads()) == set(directory.get_heads())

    def start_db_migration(self) -> None:
        """
        Runs alembic to upgrade the selected database to the current conversion head. This ensures that the database
        schema version matches the python ORM version.

        :return: Nothing
        """
        self.close_last_session()

        local_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))  # script directory
        alembic_cfg = Config(os.path.abspath(os.path.join(local_dir, "alembic.ini")))

        # Alembic has problems, when called inside a library,
        # therefore we change the current alembic folder with its absolute path
        alembic_cfg.set_main_option(
            "script_location",
            os.path.abspath(os.path.join(local_dir, alembic_cfg.get_main_option("script_location", "alembic")))
        )
        alembic_cfg.set_main_option("sqlalchemy.url", self.__connection)
        command.upgrade(alembic_cfg, "head")

    def create_new_session(self) -> Session:
        """
        Creates and returns a new session object

        :return: returns a newly created session object
        """
        self.__last_session = self.__sessionmaker()
        return self.__last_session

    def get_session(self) -> Session:
        """
        Returns the session object for the current database connection
        returns: Returns the session object for the current database connection
        """
        if self.__last_session is None:
            return self.create_new_session()
        else:
            return self.__last_session

    def close_last_session(self) -> None:
        """
        Close the actual session

        :return: Nothing
        """
        if self.__last_session is not None:
            self.__last_session.close()
            self.__last_session = None


# class AbstractDBObject(object):
class AbstractDBObject(object):
    """
    This class represents the base for all database objects. It should be treated as abstract, no object should be
    created directly!

    :param session: session object create by SQLAlchemy sessionmaker
    :type session: Session
    :param name: used to group objects by name
    :type name: str
    :param comment: additional comment
    :type comment: str
    :return: Nothing
    :raises TypeError: if session is not of type SQLAlchemy Session
    """

    id = None
    name_col = sq.Column(sq.VARCHAR(100), default="")
    comment_col = sq.Column(sq.VARCHAR(100), default="")

    def __init__(self, session: Session, name: str = "", comment: str = "") -> None:
        """
        Initialises the class
        """

        if not isinstance(session, Session):
            raise TypeError("'session' value is not of type SQLAlchemy Session!\n{} - {}".format(str(type(session)),
                                                                                                 str(Session)))

        self.__session = session
        self.comment = comment
        self.name = name

        # ABCMeta.__init__(self)

    def __repr__(self) -> str:
        return "<AbstractDBObject(name='{}', comment='{}')>".format(self.name, self.comment)

    def __str__(self) -> str:
        return "{} - {}".format(self.name, self.comment)

    @property
    def comment(self) -> str:
        """
        The additional comments for the AbstractDBObject
        """
        return self.comment_col

    @comment.setter
    def comment(self, comment: str) -> None:
        """
        see getter
        """
        comment = str(comment)
        if len(comment) > 100:
            comment = comment[:100]
        self.comment_col = comment

    def get_id(self) -> int or None:
        """
        Returns the database object id or None, if not saved / flushed to the database
        """
        return self.id

    @property
    def name(self) -> str:
        """
        The name of the AbstractDBObject
        """
        return self.name_col

    @name.setter
    def name(self, new_name: str) -> None:
        """
        see getter
        """
        new_name = str(new_name)
        if len(new_name) > 100:
            new_name = new_name[:100]
        self.name_col = new_name

    @property
    def session(self) -> Session:
        """
        The current Session object

        :raises TypeError: if session is not of an instance of Session
        """
        return self.__session

    @session.setter
    def session(self, session: Session) -> None:
        """
        see getter
        """

        if not isinstance(session, Session):
            raise TypeError("Value is not of type {} (it is {})!".format(Session, type(session)))

        self.__session = session

    # save point to db / update point
    def save_to_db(self) -> None:
        """
        Saves all changes of the object to the database

        :return: Nothing
        :raises IntegrityError: raises IntegrityError if the commit to the database fails and rolls all changes back
        """
        self.__session.add(self)
        try:
            self.__session.commit()
        except IntegrityError as e:
            # Failure during database processing? -> rollback changes and raise error again
            self.__session.rollback()
            raise IntegrityError(
                "Cannot commit changes in table, Integrity Error (double unique values?) -- {} -- " +
                "Rolling back changes...".format(e.statement), e.params, e.orig, e.connection_invalidated)
        finally:
            pass
            # self.__session.close()

    @classmethod
    def delete_from_db(cls, obj: "AbstractDBObject", session: Session):
        """
        Deletes an object from the database
        :param obj: object to delete
        :type obj; object
        :param session: SQLAlchemy Session handling the connection to the database
        :type session: Session
        :return: Nothing
        """
        session.delete(obj)
        session.commit()
        session.close()

    @classmethod
    def load_all_from_db(cls, session: Session) -> List["AbstractDBObject"]:
        """
        Returns all objects in the database connected to the SQLAlchemy Session session

        :param session: represents the database connection as SQLAlchemy Session
        :type session: Session
        :return: a list of objects
        :raises TypeError: if session is not of type SQLAlchemy Session
        """
        if not isinstance(session, Session):
            raise TypeError("'session' is not of type SQLAlchemy Session!")

        result = session.query(cls)
        result = result.order_by(cls.id).all()
        for marker in result:
            marker.session = session
        return result

    @classmethod
    def load_by_id_from_db(cls, _id: int, session: Session) -> "AbstractDBObject":
        """
        Returns the object with the given id in the database

        :param _id: Only the object with this id will be returned
        :param session: represents the database connection as SQLAlchemy Session
        :return: a single object representing the result of the database query
        :raises DatabaseRequestException: Raises DatabaseRequestException if no object was found with this id
        :raises IntegrityError: Raises IntegrityError if more than one object was found (more than one unique value)
        :raises TypeError: if session is not of type SQLAlchemy Session
        """
        if not isinstance(session, Session):
            raise TypeError("'session' is not of type SQLAlchemy Session!")

        try:
            result = session.query(cls).filter(cls.id == _id).one()
            result.session = session
            return result
        except NoResultFound:
            raise DatabaseRequestException("No result found for ID {}".format(_id))

    @classmethod
    def load_by_name_from_db(cls, name: str, session: Session) -> List["AbstractDBObject"]:
        """
        Returns all objects with the given name in the database

        :param name: Only objects or derived types with this name will be returned
        :param session: represents the database connection as SQLAlchemy Session
        :return: a list of objects representing the result of the database query
        :raises TypeError: if session is not of type SQLAlchemy Session
        """
        if not isinstance(session, Session):
            raise TypeError("'session' is not of type SQLAlchemy Session!")

        result = session.query(cls).filter(cls.name_col == name).order_by(cls.id).all()
        for obj in result:
            obj.session = session
        return result
